import numpy as np
import pandas as pd
import sys
import networkx as nx
import matplotlib.pyplot as plt
from nltk.corpus import wordnet as wn

def main(input_file):
    hypernymy_ds = './processed-webisalod-pairs.txt'
    root_node = 'entity'
    G = nx.DiGraph()
    df = pd.read_csv(hypernymy_ds, names=['hyponym', 'hypernym', 'score'], engine='python')
    df_count = df['hypernym'].value_counts(ascending=False)
    inputs = np.array([])
    # extract the input classes or entities
    # save them in an array
    with open(input_file, 'r', encoding='utf8') as fin:
        for line in fin.readlines():
            input = line.strip()
            if input != '':
                inputs = np.append(inputs, input)
    sorted_inputs = sort_inputs(inputs, df_count)
    # loop through the inputs
    for (i, input) in enumerate(sorted_inputs):
        # process taxonomy recursively
        processTaxonomy(input, df, inputs, G, root_node)
        
    # draw taxonomy
    G = G.reverse()
    '''
    # fix root node to the top right position
    fixed_root_pos = {root_node:(5,5)}
    fixed_root_nodes = fixed_root_pos.keys()
    '''
    pos = nx.spring_layout(G)
    nx.draw_networkx(G, pos)
    # plot in debug mode
    # plt.pause(1)
    # plt.show(block=True)
    # plot in run mode
    plt.show()
   
    

def sort_inputs(inputs, df_count):
    sorted_inputs = [key for key, value in df_count.items() if key in inputs][::-1]
    # fill with the rest of inputs if they are not found in hypernymy ds
    if len(inputs) != len(sorted_inputs):
        for input in inputs:
            if input not in sorted_inputs:
                sorted_inputs.append(input)
    return sorted_inputs

def processTaxonomy(hyponym: str, df: pd.DataFrame, inputs: np.array, G: nx.DiGraph, root_node='entity'):
    max_nodes = 45
    # search hypernym relation to the hyponym
    # remove actual inputs for the reference
    safe_inputs = inputs[inputs != hyponym]
    excluded_hypernyms =  []
    if hyponym in G:
        # search hyponyms for the hyponym and add to excluded_hypernyms, break infinite loop
        hyponyms = [key for key, value in nx.shortest_path_length(G, hyponym).items() if value >= 0]
        excluded_hypernyms =  hyponyms
    hypernym = searchHypernym(hyponym, df, safe_inputs, root_node, excluded_hypernyms)
    print(hyponym + ' hypernym: ' + hypernym)
    G.add_node(hyponym)
    G.add_node(hypernym)
    G.add_edge(hypernym, hyponym)
    depth = np.max([value for key, value in nx.shortest_path_length(G, hypernym).items()])
    if len(G.nodes) >= max_nodes:
        # stop creating nodes and link all roots to entity
        roots = [r for r,d in G.in_degree() if d==0]
        G.add_node(root_node)
        for root in roots:
            G.add_edge(root_node, root)
    elif depth < 4 and hypernym != root_node:
        processTaxonomy(hypernym, df, inputs, G, root_node)
    else:
        G.add_node(root_node)
        G.add_edge(root_node, hypernym)

def searchHypernym(hyponym, df, inputs, root_node='entity', excluded_hypernyms = []):
    # general constrains
    # get highest top 10 scoring hypernyms
    # exclude hypernyms
    hyponym_relations: pd.DataFrame = df[(df.hyponym == hyponym) & (df.hypernym.isin(excluded_hypernyms) == False)].sort_values(by='score', ascending=False, na_position='first').head(10)
    # get a hypernym within top 10 that is equal to other line of the hyponym
    relations = hyponym_relations[hyponym_relations['hypernym'].isin(inputs)]
    if len(relations) > 0:
        relation = relations.iloc[0]['hypernym']
    else: # get a hypernym within top 20 that lexically matches with the hyponym (substring of it)
        matches = [match for match in hyponym_relations['hypernym'] if match in hyponym and match != hyponym]
        if len(matches) > 0:
            relation = matches[0]
        elif len(hyponym_relations) > 0: # get the top scorer
            relation = hyponym_relations.iloc[0]['hypernym']
        else: # if there are no hypernyms for the hyponym
            # search in word embeddings and interesect with an hypernym in the webisalod
            synsets = [synset for synset in wn.synsets(hyponym)]
            if len(synsets) > 0:
                hypernyms = [hypernym for hypernym in synsets[0].hypernyms()]
                if len(hypernyms) > 0:
                    relation = hypernyms[0].lemma_names()[0]
                else:
                    relation = root_node
            else: # if nothing is found, hypernym set to root node 
                relation = root_node
    return relation
   
'''
main function
'''
if __name__ == '__main__':
    if len(sys.argv) != 2:
        raise ValueError('Expected exactly 1 argument: input file')
    main(sys.argv[1])