import sys
from typing import Optional
from urllib.parse import unquote
import networkx as nx
import matplotlib.pyplot as plt
import wn
from wn import Synset
from graph import TreeGenerator
import os

root_node_synonyms = ["entity", "thing"]
en = wn.Wordnet('oewn:2021')

relation_threshold = 0.55
min_path_length = 4
max_nodes_count = 50


class WordPathNode:
    def __init__(self, word: str, vital: bool, synset: Optional[Synset]):
        self.word = word
        self.vital = vital
        self.synset = synset

    def get_word(self):
        return self.word

    def get_vital(self):
        return self.vital

    def get_synset(self):
        return self.synset


def load_webisalod_pairs(webisalod_file_path):
    fin = open(webisalod_file_path, 'r', encoding='utf8')
    clean_data = []
    for line in fin.readlines():
        line_comps = line.strip().split("\t")
        pair = line_comps[0].split(";")
        similarity = float(line_comps[1])
        for i in range(len(pair)):
            pair[i] = unquote(pair[i]) \
                .replace("_", " ") \
                .replace("+", " ") \
                .strip()
        clean_data.append({
            'hyponym': pair[0],
            'hypernym': pair[1],
            'similarity': similarity
        })

    # Hyponym = child
    # Hypernym = parent
    # Hyponym outer, Hypernym inner
    search_map_hypernym = {}
    search_map_hyponym = {}
    for cd in clean_data:
        hyponym = cd['hyponym']
        hypernym = cd['hypernym']
        similarity = cd['similarity']

        if hyponym in search_map_hypernym:
            inner_hypernym_map = search_map_hypernym[hyponym]
        else:
            inner_hypernym_map = {}
            search_map_hypernym[hyponym] = inner_hypernym_map

        if hypernym in search_map_hyponym:
            inner_hyponym_map = search_map_hyponym[hypernym]
        else:
            inner_hyponym_map = {}
            search_map_hyponym[hypernym] = inner_hyponym_map

        if similarity > relation_threshold:
            inner_hypernym_map[hypernym] = similarity
            inner_hyponym_map[hyponym] = similarity

    return search_map_hypernym, search_map_hyponym


def get_hypernyms_sorted_by_confidence(hyponym_child, search_map):
    if hyponym_child not in search_map:
        return {}
    hypernyms = search_map[hyponym_child]
    return {k: v for k, v in sorted(hypernyms.items(), key=lambda item: item[1], reverse=True)}


def get_webisalod_similarity(hyponym_child: str, hypernym: str) -> float:
    if hyponym_child not in webisalod_search_map_hypernym:
        return 0.0
    inner_map = webisalod_search_map_hypernym[hyponym_child]
    if hypernym not in inner_map:
        return 0.0
    return inner_map[hypernym]


def select_next_wil_hypernym(path: list[WordPathNode], input_words: list[str], is_hypernym) -> Optional[WordPathNode]:
    if len(path) == 0:
        return None
    if is_hypernym:
        search_map = webisalod_search_map_hypernym
        last_word = path[len(path) - 1].get_word()
        control_word = path[0].get_word()
    else:
        search_map = webisalod_search_map_hyponym
        last_word = path[0].get_word()
        control_word = path[1 if len(path) > 1 else 0].get_word()

    wil_hypernyms = list(get_hypernyms_sorted_by_confidence(last_word, search_map).keys())
    # Remove all words that are already part of the path
    for word_node in path:
        if word_node.get_word() in wil_hypernyms:
            wil_hypernyms.remove(word_node.get_word())

    for hypernym in wil_hypernyms:
        if get_webisalod_similarity(control_word, hypernym) < 0.4 \
                and get_webisalod_similarity(last_word, hypernym) < 0.8:
            wil_hypernyms.remove(hypernym)

    # If one of the possible hypernyms is part of our input list, choose that one
    for input_word in input_words:
        if input_word in wil_hypernyms:
            return WordPathNode(input_word, True, None)

    # If hypernym list not empty, use the hypernym with the highest confidence
    if len(wil_hypernyms) != 0:
        return WordPathNode(wil_hypernyms[0], False, None)

    return None


def get_wordnet_paths(word: str, input_words: list[str]) -> list[list[WordPathNode]]:
    wn_paths = []
    synsets = en.synsets(word, pos='n')
    for synset in synsets:
        for hyper_path in synset.hypernym_paths():
            syn_path = [WordPathNode(word, word in input_words, synset)]
            for word_synset in hyper_path:
                if len(word_synset.lemmas()) != 0:
                    lemma = word_synset.lemmas()[0]
                    syn_path.append(WordPathNode(lemma, lemma in input_words, word_synset))
            wn_paths.append(syn_path)
    return wn_paths


def get_next_hyponym(path: list[WordPathNode], input_words):
    path_length = len(path)
    if path_length == 0:
        return None
    top_word = path[0]
    synset = top_word.get_synset()
    if synset is not None:
        wn_hyponyms = synset.hyponyms()
        for wn_hyponym in wn_hyponyms:
            for lemma in wn_hyponym.lemmas():
                return WordPathNode(lemma, False, wn_hyponym)
    return select_next_wil_hypernym(path, input_words, False)


def expand_path(path: list[WordPathNode], input_words, to_size):
    path_length = len(path)
    if path_length != 0 and path_length >= to_size:
        return path
    next_hyponym = get_next_hyponym(path, input_words)
    while next_hyponym is not None and path_length < to_size:
        path.insert(0, next_hyponym)
        path_length = len(path)
        next_hyponym = get_next_hyponym(path, input_words)
    return path


def gen_path_till_root(leaf_word, input_words) -> list[list[WordPathNode]]:
    wil_path = []
    wn_paths = get_wordnet_paths(leaf_word, input_words)

    if len(wn_paths) == 0:
        wil_path.append(WordPathNode(leaf_word, leaf_word in input_words, None))
        next_element = select_next_wil_hypernym(wil_path, input_words, True)
        while next_element is not None:
            wn_paths = get_wordnet_paths(next_element.get_word(), input_words)
            if len(wn_paths) != 0:
                break
            else:
                wil_path.append(next_element)
                next_element = select_next_wil_hypernym(wil_path, input_words, True)

    found_paths = []
    if len(wn_paths) != 0:
        # If wordnet path(s) found. Prepend all WIL-path-words that have been found before
        for wn_path in wn_paths:
            full_path = list(wil_path)
            full_path.extend(wn_path)
            found_paths.append(full_path)
    else:
        found_paths = [wil_path]

    for path in found_paths:
        if len(path) != 0:
            entity_word = "entity"
            if path[len(path) - 1].get_word() != entity_word:
                path.append(WordPathNode(entity_word, entity_word in input_words, None))

    for index in range(len(found_paths)):
        path = found_paths[index]
        if len(path) != 0 and len(path) < min_path_length:
            found_paths[index] = expand_path(path, input_words, min_path_length)
        found_paths[index].reverse()

    return found_paths


def construct_taxonomy_tree(input_file_path):
    fin = open(input_file_path, 'r', encoding='utf8')
    input_words = [x.strip() for x in fin.readlines()]
    input_words = [word for word in input_words if word]
    fin.close()

    word_paths = {}
    for word in input_words:
        word_paths[word] = gen_path_till_root(word, input_words)

    word_paths_for_gen = {}
    for word, paths in word_paths.items():
        word_paths_for_gen[word] = []
        for path in paths:
            path_for_gen = []
            for wpn in path:
                path_for_gen.append(wpn.get_word())
            word_paths_for_gen[word].append(path_for_gen)
    tree_gen = TreeGenerator(word_paths_for_gen)
    root = tree_gen.generate()

    plot_dir = "graphs"
    if not os.path.exists(plot_dir):
        os.makedirs(plot_dir)
    original_plot_filepath = "graphs/original.png"
    if os.path.exists(original_plot_filepath):
        os.remove(original_plot_filepath)
    plt.figure()
    graph = root.plot_tree()
    nx.draw(graph, with_labels=True)
    plt.savefig(original_plot_filepath)
    #plt.show()

    truncated_success = True
    while root.count_nodes() > max_nodes_count and truncated_success:
        truncated_success = root.truncate_one(min_path_length, input_words)

    nodes_50_plot_filepath = "graphs/max-50-nodes.png"
    if os.path.exists(nodes_50_plot_filepath):
        os.remove(nodes_50_plot_filepath)
    plt.figure()
    graph = root.plot_tree()
    nx.draw(graph, with_labels=True)
    plt.savefig(nodes_50_plot_filepath)
    #plt.show()

    truncated_success = True
    while truncated_success:
        truncated_success = root.truncate_one(min_path_length, input_words)

    minimal_plot_filepath = "graphs/minimal-nodes.png"
    if os.path.exists(minimal_plot_filepath):
        os.remove(minimal_plot_filepath)
    plt.figure()
    graph = root.plot_tree()
    nx.draw(graph, with_labels=True)
    plt.savefig(minimal_plot_filepath)
    #plt.show()


if __name__ == '__main__':
    if len(sys.argv) != 2:
        raise ValueError('Expected exactly 1 argument: input file')
    webisalod_search_map_hypernym, webisalod_search_map_hyponym = load_webisalod_pairs('webisalod-pairs.txt')
    construct_taxonomy_tree(sys.argv[1])
