import sys
from urllib.request import unquote

import networkx as nx
import matplotlib.pyplot as plt
import pydot
from networkx.drawing.nx_pydot import graphviz_layout

#pip install pydot-ng
#brew install graphviz


def parse_webisalod_line(full_graph,line,threshhold):
    split_line= line.split('\t')
    confidence = float(split_line[1])
    if(confidence>=threshhold):
        hyponym = parse_webisalod_string(split_line[0].split(' ')[0])
        hypernym = parse_webisalod_string(split_line[0].split(' ')[2])
        if hyponym == hypernym:
            return
        if hyponym in full_graph:
            full_graph[hyponym].add((hypernym,confidence))
        else:
            full_graph[hyponym]=set({(hypernym,confidence)})

def parse_webisalod_string(string):
    if string[-1]=='_':
        string = string[:-1]
    if string[0]=='_':
        string = string[1:]
    string = string.replace('_',' ')
    string = string.replace('+',' ')
    return unquote(string)

def build_webisalod_graph(threshhold=0.6):
    full_graph={}
    with open('webisalod-pairs.txt') as file:
        for line in file:
            parse_webisalod_line(full_graph,line,threshhold)
    return full_graph

class taxonomy_node:
    def __init__(self,name,is_root=False,is_original_term=False):
        self.name=name
        self.descendants={}
        self.is_root=is_root
        self.is_original_term = is_original_term
        self.is_leaf = is_original_term
        self.ancestor_node=None
    def has_ancestor(self,name):
        if (self.name==name):
            return True;
        if self.ancestor_node:
            return self.ancestor_node.has_ancestor(name)
        return False;
    def get_height(self):
        if self.descendants:
            max_val=0
            for descendant in self.descendants.values():
                max_val=max(max_val,descendant.get_height())
            return max_val+1
        return 0
    def prune_node(self,nodes):
        if self.is_original_term:
            branch_names=list(self.descendants.keys())
            for name in branch_names:
                if descendant.prune_node(nodes):
                    return True
            return False
        #prune
        for name,descendant in self.descendants.items():
            self.ancestor_node.descendants[name]=descendant
            descendant.ancestor_node=self.ancestor_node
        del self.ancestor_node.descendants[self.name]
        del nodes[self.name]
        return True

    def prune_node_preserve_height(self,nodes,height):
        has_pruned_node=False
        has_sufficiently_long_branch=False
        sufficiently_long_branch=""
        branch_names=list(self.descendants.keys())
        for name in branch_names:
            if nodes[name].get_height()>=height:
                if has_sufficiently_long_branch:
                    has_pruned_node = nodes[name].prune_node(nodes)
                else:
                    has_sufficiently_long_branch=True
                    sufficiently_long_branch=name
            else:
                has_pruned_node = nodes[name].prune_node(nodes)
            if (has_pruned_node):
                break

        if not has_pruned_node:
            if has_sufficiently_long_branch:
                return nodes[sufficiently_long_branch].prune_node_preserve_height(nodes,height-1)
            else:
                return False
        else:
            return True


class taxonomy_builder:
    def __init__(self,max_items=50,minimum_height=4):
        self.hypernym_graph = build_webisalod_graph()
        self.nodes = {}
        self.root=None
        self.max_items=max_items
        self.minimum_height=minimum_height

    def initialize_taxonomy_nodes(self,file_name):
        with open(file_name) as file:
            for line in file:
                stripped_line = line.strip()
                if stripped_line:
                    self.nodes[stripped_line]=taxonomy_node(stripped_line,is_root=True,is_original_term=True)
    def build_taxonomy_from_file(self,file_name):
        self.initialize_taxonomy_nodes(file_name)
        #Eventually break into own method
        for i in range(self.minimum_height):
            nodes_names=list(self.nodes.keys())
            for name in nodes_names:
                if self.nodes[name].is_root and self.nodes[name].get_height()<self.minimum_height:
                    self.link_ancestor(name,self.find_ancestor(name))
        nodes_names=[]
        for name,node in self.nodes.items():
            if node.is_root:
                nodes_names.append(name)
        if len(nodes_names)>1:
            for name in nodes_names:
                self.link_ancestor(name,"_entity_")
            self.root=self.nodes["_entity_"]
        else:
            self.root=self.nodes[nodes_names[0]]
        self.prune_taxonomy()
    def link_ancestor(self,name,ancestor_name):
        if not ancestor_name:
            return
        if ancestor_name in self.nodes:
            self.nodes[ancestor_name].is_leaf=False
        else:
            new_ancestor=taxonomy_node(ancestor_name,is_root=True)
            self.nodes[ancestor_name] = new_ancestor
        self.nodes[name].is_root=False
        self.nodes[name].ancestor_node = self.nodes[ancestor_name]
        self.nodes[ancestor_name].descendants[name] = self.nodes[name]
    def find_ancestor(self,name):
        max_existing_ancestor=""
        max_existing_confidence=0
        max_new_ancestor=""
        max_new_confidence=0
        if name not in self.hypernym_graph:
            return ""
        for ancestor in self.hypernym_graph[name]:
            ancestor_name = ancestor[0]
            ancestor_confidence = ancestor[1]
            if ancestor_confidence<=max_existing_confidence and ancestor_confidence<=max_new_confidence:
                continue
            if ancestor_name in self.nodes:
                if not self.nodes[ancestor_name].has_ancestor(name):
                    #check that the opposite relationship is not better
                    reverse_confidence=0
                    if ancestor_name in self.hypernym_graph:
                        for hypernym in self.hypernym_graph[ancestor_name]:
                            if hypernym[0]==name:
                                reverse_confidence = hypernym[1]
                                break;
                    if reverse_confidence<=ancestor_confidence:
                        max_existing_ancestor=ancestor_name
                        max_existing_confidence=ancestor_confidence

            elif (ancestor_confidence>max_new_confidence):
                max_new_ancestor=ancestor_name
                max_new_confidence=ancestor_confidence
        if max_existing_confidence>0:
            return max_existing_ancestor
        return max_new_ancestor

    def prune_taxonomy(self):
        has_pruned_node=True
        while(has_pruned_node):
            has_pruned_node=False
            branch_names=list(self.root.descendants.keys())
            for name in branch_names:
                if self.nodes[name].get_height()>=self.minimum_height+1:
                    has_pruned_node = self.nodes[name].prune_node(self.nodes)
                if (has_pruned_node):
                    break

        has_pruned_node=True
        while(len(self.nodes)>self.max_items and has_pruned_node):
            has_pruned_node=False
            branch_names=list(self.root.descendants.keys())
            for name in branch_names:
                if self.nodes[name].get_height()>=self.minimum_height:
                    has_pruned_node = self.nodes[name].prune_node(self.nodes)
                if (has_pruned_node):
                    break
        has_pruned_node=True
        while (len(self.nodes)>self.max_items and has_pruned_node):
            has_pruned_node= self.root.prune_node_preserve_height(self.nodes,self.minimum_height)




    def display_taxonomy(self):
        plt.figure(figsize=(15, 7))
        G = nx.DiGraph()
        for node in self.nodes.values():
            if not node.is_root:
                G.add_edges_from([(node.name,node.ancestor_node.name)])
        color_map = []
        for node in G:
            if self.nodes[node].is_original_term:
                color_map.append('red')
            else:
                color_map.append('green')
        G = nx.relabel_nodes(G, {'_entity_':'entity'})

        pos = graphviz_layout(G, prog="twopi")
        nx.draw_networkx_nodes(G, pos,node_color = color_map,node_size = 100)
        nx.draw_networkx_labels(G, pos)
        nx.draw_networkx_edges(G, pos,edge_color='blue',arrows=True)
        plt.show()




'''
main function
'''

if __name__ == '__main__':
    if len(sys.argv) != 2:
        raise ValueError('Expected exactly 1 argument: input file')

    tb = taxonomy_builder()
    tb.build_taxonomy_from_file(sys.argv[1])
    if tb.root.get_height()<tb.minimum_height: #Should not reasonably happen, but is a safety measure
        tb.hypernym_graph = build_webisalod_graph(threshhold=0.4)
        tb.build_taxonomy_from_file(sys.argv[1])
    if tb.root.get_height()<tb.minimum_height: #Should not reasonably happen, but is a safety measure
        tb.hypernym_graph = build_webisalod_graph(threshhold=0.0)
        tb.build_taxonomy_from_file(sys.argv[1])
    tb.display_taxonomy()
