import sys
from treelib import Node, Tree
import igraph as ig
import numpy as np

'''
reads input file with words to be used as entities, stores as list
'''
def read_input(input_file_path):
    entities = [] # store entities from input file
    with open (input_file_path, 'r', encoding='utf-8') as in_file:
        for line in in_file:
            if line.strip():    # remove empty lines
                entities.append(line.strip())
    return entities

'''
reads webisalod corpus, stores as dict of dict, mapping from hyponym to hypernym to score
'''
def read_webisalod(corpus_path):
    hypo_to_hyper_to_score = {}
    with open (corpus_path, 'r', encoding='utf-8') as hypernyms:
        for line in hypernyms:
            hyponym = line.split(';')[0].strip()
            hypernym = line.split(';')[1].split()[0].strip()
            score = line.split(';')[1].split()[1].strip()
            if hyponym == hypernym:     # if hyponym equals hypernym, skip, do not add to dict
                continue
            if hyponym in hypo_to_hyper_to_score:
                hypo_to_hyper_to_score[hyponym][hypernym] = score
            else:
                hypo_to_hyper_to_score[hyponym] = {}
                hypo_to_hyper_to_score[hyponym][hypernym] = score
    return hypo_to_hyper_to_score

'''
takes entity in webisalod format, returns a word 
'''
def ent_to_word(e):
    w = ""
    if '+' in e:
        w = e.replace('+',' ').replace('_',' ').strip()
    else:
        w = e.replace('_',' ').strip()
    return w

'''
takes word, returns it in entity format of webisalod 
! does not deal adequately with words with '+' inside
'''
def word_to_ent(w):  
    if ' ' in w:     # get word in search format
        e = w.replace(' ','_')
        e += '_'  
    else:
        e = '_' + w + '_' 
    return e

'''
get all possible hypernyms of a node (in word form) from corpus
'''
def get_all_hypernyms(node, mapping_dict):
    ent = node
    all_parents = []
    all_scores = []
    if ent in mapping_dict:
        for hyper_ent in mapping_dict[ent]:   # go through possible hypenyms
            all_parents.append(hyper_ent)
            all_scores.append(float(mapping_dict[ent][hyper_ent]))
    else:
        print("Entity not found in corpus: ")
        print(ent)
    
    return all_parents, all_scores


'''
construct the tree from input words
'''
def build_tree(input_entities, hypo_to_hyper_to_score):
    tree = Tree()
    tree.create_node("ENTITY", "entity")  # create root node
    child_to_possible_parents = {}
    child_to_all_parents = {}
    child_to_all_parent_scores = {}

    for word in input_entities:
        ent = word_to_ent(word)
        if ent in hypo_to_hyper_to_score:

            possible_parents = []
            parent_scores = []
            all_parents = []
            all_scores = []

            score= 0
            highest_score_hyp = ""  # store hypernym with highest score which is in list
            for hyper_ent in hypo_to_hyper_to_score[ent]:   # go through possible hypenyms
                hyper = ent_to_word(hyper_ent)
                all_parents.append(hyper_ent)
                all_scores.append(float(hypo_to_hyper_to_score[ent][hyper_ent]))
                if hyper in input_entities:
                    possible_parents.append(hyper_ent)
                    parent_scores.append(hypo_to_hyper_to_score[ent][hyper_ent])
                    if float(hypo_to_hyper_to_score[ent][hyper_ent]) > score:
                        score = float(hypo_to_hyper_to_score[ent][hyper_ent])
                        highest_score_hyp = hyper_ent
            
            child_to_possible_parents[ent] = possible_parents
            child_to_all_parents[ent] = all_parents
            child_to_all_parent_scores[ent] = all_scores

            
            if not tree.get_node(ent):  # check if word is already in tree; if it is, no need to add to tree
                # take hypernym with highest score which is in list and try to make parent
                if len(highest_score_hyp) > 0:  # if hypernym found
                    if not tree.get_node(highest_score_hyp):# if parent does not exist, create       
                        tree.create_node(highest_score_hyp, highest_score_hyp, parent="entity")
                        tree.create_node(ent, ent, parent=highest_score_hyp)  
                    else:
                        # if parent node already exists, add to it
                        tree.create_node(ent, ent, parent=highest_score_hyp)
                else: # if hypernym not found, append node to root
                    tree.create_node(ent, ent, parent="entity")
   
        else:
            continue

    # sibling check
    for i in tree.all_nodes():
        if i.is_leaf():
            highest_sibling_score = 0.4 # start with 0.4 boundary
            highest_sibling = ""
            for sib in tree.siblings(i.identifier):
                if sib.identifier in child_to_possible_parents[i.identifier]:
                    if float(hypo_to_hyper_to_score[i.identifier][sib.identifier]) > highest_sibling_score:
                        highest_sibling_score = float(hypo_to_hyper_to_score[i.identifier][sib.identifier])
                        highest_sibling = sib.identifier
            if len(highest_sibling) > 0:
                tree.move_node(i.identifier, highest_sibling)



    # depth check
    for path in tree.paths_to_leaves():

        if len(path) < 4: # depth check
            
            last_node = path[1]
            for i in range(4-len(path)):

                highest_node = last_node

                if highest_node in hypo_to_hyper_to_score:

                    # get indices of hypers with highest values
                    all_parents, all_scores = get_all_hypernyms(highest_node, hypo_to_hyper_to_score_mapping)
                    all_scores_np = np.array(all_scores)
                    all_parents_np = np.array(all_parents)
                    ind = np.argpartition(all_scores_np, -1)[-1:]
                    ind_reverse = ind[::-1]
                    top_scores = all_scores_np[ind]
                    top_parents = all_parents_np[ind]


                    if not tree.get_node(top_parents[0]):# if parent does not exist, create       
                        tree.create_node(top_parents[0], top_parents[0], parent="entity")
                        tree.move_node(highest_node, top_parents[0])  
                    else:
                        # if parent node already exists, add to it
                        tree.move_node(highest_node, top_parents[0])

                    last_node = top_parents[0]


    # tree.show()
    return tree

'''
plots the constructed tree using Igraph
'''
def plot_tree(tree):

    # make labels for tree drawing
    node_to_idx = {}    # map tree identifiers to indices
    idx = 0
    labels = []     # make labels for drawing tree
    for node in tree.all_nodes():
        node_to_idx[node.identifier] = idx
        idx+=1
        labels.append(node.identifier)

    # collect tree edges for drawing
    edges_all = []
    for path in tree.paths_to_leaves():
        edges_path = ([(x,y) for x,y in zip(path, path[1:])])   # make tuples for tree printing
        for edge in edges_path:
            edge_list = [node_to_idx[vertex] for vertex in edge]
            edge = tuple(edge_list[::-1] )
            if edge not in edges_all:
                edges_all.append(edge)

    # draw tree    
    g = ig.Graph(n = tree.size(), directed=True)
    g.add_edges(edges_all)
    g.vs["label"] = [ent_to_word(label) for label in labels]
    ig.plot(g, layout="kk", vertex_color='pink', vertex_label_size=12, vertex_size=33).show()

'''
main function
'''
if __name__ == '__main__':

    if len(sys.argv) != 2:
        raise ValueError('Expected exactly 1 argument: input file')
    
    print("Reading input...")
    webisalod_path = './lab04_material/webisalod-pairs.txt'  # set path to WEBISALOD file here
    entity_input_list = read_input(sys.argv[1])
    print("Reading WEBISALOD corpus... (takes around 40sec)")
    hypo_to_hyper_to_score_mapping = read_webisalod(webisalod_path)
    print("Building tree...")
    tree = build_tree(entity_input_list, hypo_to_hyper_to_score_mapping)
    plot_tree(tree)
    print("Done!")